import { BadArgumentError } from '@/utils';
import type { DenotableValueType } from '.';

export type Signature = Array<Array<DenotableValueType | Signature>>;

export const primitiveOperationSignatures: Record<string, Signature> = {
  '+': [
    ['int', 'int', 'int'],
    ['real', 'real', 'real'],
    ['int', 'real', 'real'],
    ['real', 'int', 'real'],
  ],
  '-': [
    ['int', 'int', 'int'],
    ['real', 'real', 'real'],
    ['int', 'real', 'real'],
    ['real', 'int', 'real'],
  ],
  '*': [
    ['int', 'int', 'int'],
    ['real', 'real', 'real'],
    ['int', 'real', 'real'],
    ['real', 'int', 'real'],
  ],
  '/': [
    ['int', 'int', 'int'],
    ['real', 'real', 'real'],
    ['int', 'real', 'real'],
    ['real', 'int', 'real'],
  ],
  '%': [['int', 'int', 'int']],
  '**': [
    ['int', 'int', 'int'],
    ['real', 'real', 'real'],
    ['int', 'real', 'real'],
    ['real', 'int', 'real'],
  ],
  '>>': [['int', 'int', 'int']],
  '<<': [['int', 'int', 'int']],
  '&': [['int', 'int', 'int']],
  '|': [['int', 'int', 'int']],
  '^': [['int', 'int', 'int']],
  '==': [
    ['int', 'int', 'int'],
    ['real', 'real', 'int'],
    ['string', 'string', 'int'],
  ],
  '!=': [
    ['int', 'int', 'int'],
    ['real', 'real', 'int'],
    ['string', 'string', 'int'],
  ],
  '<': [
    ['int', 'int', 'int'],
    ['real', 'real', 'int'],
  ],
  '>': [
    ['int', 'int', 'int'],
    ['real', 'real', 'int'],
  ],
  '<=': [
    ['int', 'int', 'int'],
    ['real', 'real', 'int'],
  ],
  '>=': [
    ['int', 'int', 'int'],
    ['real', 'real', 'int'],
  ],
  '&&': [['int', 'int', 'int']],
  '||': [['int', 'int', 'int']],
  '!': [['int', 'int']],
};

export const getResultingType = (
  opr: string,
  types: DenotableValueType[],
): DenotableValueType | Signature => {
  const signature = primitiveOperationSignatures[opr];
  if (!signature) {
    throw new BadArgumentError(`Invalid operation: ${opr}`);
  }

  const resultingType = signature.find(sig => {
    if (sig.length !== types.length) {
      return false;
    }

    return sig.every((type, i) => {
      if (Array.isArray(type)) {
        return getResultingType(
          opr,
          types.map((t, j) => (i === j ? t : 'int')),
        );
      }
      return type === 'int' || type === types[i];
    });
  });

  if (!resultingType) {
    throw new TypeError(
      `Invalid types for operation ${opr}: ${types.join(', ')}`,
    );
  }

  return resultingType as DenotableValueType | Signature;
};
